package tetris.game;

import java.util.*;
import java.awt.Color;

public class GameModel {
	private Color[][] field;
	private ArrayList<GameBlock> blocks;
	private GameListener glistener;

	GameBlock activeBlock;
	int ax, ay; // Offset coordinates of the current active block

	public GameModel(GameListener gl, int w, int h) {
		glistener = gl;

		field = new Color[h][w];
		for (int i = 0; i < h; ++i) {
			for (int j = 0; j < w; ++j) {
				field[i][j] = Color.black;
			}
		}

		blocks = new ArrayList<GameBlock>();

		blocks.add(new GameBlock(new boolean[][]
				{{true, true},
				 {true, true}}, Color.yellow));
		blocks.add(new GameBlock(new boolean[][]
				{{ true,  true,  true},
				 {false,  true, false}}, Color.magenta));
		blocks.add(new GameBlock(new boolean[][]
				{{ true,  true,  true},
				 {false, false,  true}}, Color.blue));
		blocks.add(new GameBlock(new boolean[][]
				{{ true,  true,  true},
				 { true, false, false}}, Color.orange));
		blocks.add(new GameBlock(new boolean[][]
				{{false,  true,  true},
				 { true,  true, false}}, Color.green));
		blocks.add(new GameBlock(new boolean[][]
				{{ true,  true, false},
				 {false,  true,  true}}, Color.red));
		blocks.add(new GameBlock(new boolean[][]
				{{true, true, true, true}}, Color.cyan));
	}

	public Color[][] getField() {
		return field;
	}

	public GameBlock randomBlock() {
		Random r = new Random();
		return blocks.get(r.nextInt(blocks.size())).clone();
	}

	public GameBlock getActiveBlock() {
		return activeBlock;
	}

	public int getActiveX() {
		return ax;
	}

	public int getActiveY() {
		return ay;
	}

	public void setActiveBlock(GameBlock b) {
		activeBlock = b;
		ax = field[0].length / 2;
		ay = 0;
	}
	
	// Test if the active block can be rotated left, and do it if
	// there is space then return true; otherwise return false
	public boolean rotateActiveLeft() {
		// If the rotated height would reach beyond the bottom, don't even try
		if (ay + activeBlock.shape[0].length > field.length) {
			return false;
		}

		// Store the old shape and offsets, just in case
		boolean[][] oldShape = activeBlock.shape;
		int oy = ay;
		int ox = ax;

		// Perform the rotation
		boolean[][] newShape = new boolean[oldShape[0].length][oldShape.length];
		for (int i = oldShape.length - 1 ; i >= 0 ; --i) {
			for (int j = 0 ; j < oldShape[i].length ; ++j) {
				newShape[j][oldShape.length - 1 - i] = oldShape[i][j];
			}
		}
		activeBlock.shape = newShape;
		
		// The long block needs a bit of adjustment
		if (activeBlock.color == Color.cyan) {
			if (oldShape.length > newShape.length) {
				ax -= 1;
			}
			else {
				ax += 1;
			}
		}
		
		// If the new shape is hitting anything, restore old shape and fail
		if (!isValidMove(0, 0)) {
			activeBlock.shape = oldShape;
			ay = oy;
			ax = ox;
			return false;
		}

		return true;
	}

	public boolean rotateActiveRight() {
		// ToDo: Test if the active block can be rotated right, and do it if
		// there is space then return true; otherwise return false
		return false;
	}

	// Test if the active block can be moved down 1, and do it if
	// there is space then return true; otherwise return false
	public boolean moveActiveDown() {
		if (ay+activeBlock.shape.length >= field.length) {
			return false;
		}
		if (!isValidMove(0, 1)) {
			return false;
		}
		++ay;
		return true;
	}

	public boolean moveActiveLeft() {
		if (!isValidMove(1, 0)) {
			return false;
		}
		int width = field[0].length;
		for (int op = 0 ; op < field.length ; ++op) {
			Color z = field[op][0];
			for (int right = width ; right >= 1 ; --right) {
				field[op][width-right] = field[op][(width-right+1)%width];
			}
			field[op][width-1] = z;
		}
		return true;
	}

	public boolean moveActiveRight() {
		if (!isValidMove(-1, 0)) {
			return false;
		}
		int width = field[0].length;
		for (int op = 0 ; op < field.length ; ++op) {
			Color z = field[op][0];
			for (int right = 1 ; right <= width ; ++right) {
				field[op][(width-right+1)%width] = field[op][width-right];
			}
			field[op][1] = z;
		}
		return true;
	}

	public void settleActive() {
		while (moveActiveDown()) {
			// Empty on purpose...
			// We just want to let the block move down until it can't move any more.
		}

		// Add the active block to the field, then set it null
		for (int y = 0 ; y < activeBlock.shape.length ; ++y) {
			for (int x = 0 ; x < activeBlock.shape[y].length ; ++x) {
				if (activeBlock.shape[y][x] == true) {
					field[ay+y][ax+x] = activeBlock.color;
				}
			}
		}
		activeBlock = null;

		// Check for filled rows and signal Display if one is about to be removed, then remove it
		for (int y=0 ; y < field.length ; ++y) {
			// Assume there is no black in the row
			boolean black = false;
			for (int x = 0 ; x < field[y].length ; ++x) {
				if (field[y][x] == Color.black) {
					// We found black, so this row is not removable
					black = true;
					break;
				}
			}
			if (black == false) {
				// Signal Display that the row is about to be removed
				glistener.removeRowEvent(y);
				// Move all rows down one
				for (int r = y ; r > 0 ; --r) {
					field[r] = field[r-1].clone();
				}
				// Set top row to black
				for (int l = 0 ; l < field[y].length ; ++l) {
					field[0][l] = Color.black;
				}
			}
		}

		// Check for game-over state and signal Display if so
		for (int l = 0 ; l < field[0].length ; ++l) {
			if (field[0][l] != Color.black) {
				glistener.gameOverEvent();
			}
		}

		// Signal Display that the block has been used
		glistener.blockUsedEvent();
	}
	
	private boolean isValidMove(int dx, int dy) {
		for (int i = 0; i < activeBlock.shape.length; i++) {
			for (int j = 0; j < activeBlock.shape[i].length; j++) {
				if (activeBlock.shape[i][j] == true) {
					if (field[(ay+i+dy)%field.length][(ax+j+dx)%field[0].length] != Color.black) {
						return false;
					}
				}
			}
		}
		return true;
	}
}
